import numpy as np
import numpy.random as npr
import pandas as pd
import logging


class Subsampler:
    
    def __init__(self, df_data, columns_keep_ratio=[], allowed_deviation=.2, allow_zero_occurrence=False):

        # df_ratio = df_ratio.dropna(subset=[column_target])  # remove entries in target column with NaN
        
        self.df_data = df_data.copy()  # complete input data
        self.df_ratio = self.df_data[columns_keep_ratio]  # the dataframe with the columns in which to keep the ratio
        self.allowed_deviation = allowed_deviation  # the allowed deviation from the ratio in the subsampled dataframe in the columns 'columns_keep_ratio' from 0 to 1
        
        # get the categorical columns to create dummy columns for ratio comparison. bool columns will not be in the categorical columns and will be converted to integer columns in the preprocessing
        self.categorical_columns = self.df_ratio.columns[~self.df_ratio.columns.isin(self.df_ratio._get_numeric_data().columns)]
        
        # df to compare mean deviation against when subsampling
        self.df_mean_orig = None  # will be set in preprocessing in "_make_mean_reference"

        # attributes to be filled by methods
        self._test_df = None  # contains test data, created by 'extract_test' if test_size > 0
        
        self._deviation_list = []
        
        # whether to allow 0 occurrences of a value in subsampled data, e.g. complete removal of this value due to subsampling
        self.allow_zero_occurrence = allow_zero_occurrence
        
        self._column_separator = '$$$$$$&&&&&&'
        
        # init steps
        self._preprocess()
        
        
    def _bool_to_int_columns(self):
        # normalize columns
        for col in self.df_ratio:
            if self.df_ratio[col].dtype in ['bool']:
                self.df_ratio.loc[:, col] = self.df_ratio[col].map(int)
                
        
    def _normalize_numerical_columns(self):
        # normalize columns
        for col in self.df_ratio:
            if self.df_ratio[col].dtype in ['float64', 'int64']:
                self.df_ratio.loc[:, col] = self.column_normalize(self.df_ratio[col])
                
                
    def _make_mean_reference(self):
        df = self.df_ratio
        
        if len(self.categorical_columns):
            df = df.drop(self.categorical_columns, axis=1).join(pd.get_dummies(df[self.categorical_columns],  prefix_sep=self._column_separator))
            
        self.df_mean_orig = df.mean()
        
        
    def _preprocess(self):
        # convert bool columns to numerical
        self._bool_to_int_columns()
        # normalize numerical columns
        self._normalize_numerical_columns()
        # important to make mean reference after normalization
        self._make_mean_reference()
            
    
    def extract_test(self, test_size=.2, random_state=101, raise_exception=False):
        """_summary_

        Args:
            test_size (float, optional): percentage of datapoints in test dataset which will be excluded from subsampling. Defaults to .2.
            random_state (int, optional): the random seed. Defaults to 101.

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        
        if type(test_size) not in [int, float] or not 1 >= test_size >= 0:
            raise ValueError('Parameter "test_size" has to be a value in the range from 0 to 1.')
        
        # return empty dataframe with columns 
        if test_size == 0:
            return pd.DataFrame(columns=self.df_data.columns)
        # subsample with equal data distributions
        self._test_df = self.subsample(
            test_size, random_state=random_state, raise_exception=raise_exception)
        if self._test_df is None:
            return None
        
        # remove test data points from df_data to exclude it in the subsampling process
        self.df_ratio = self.df_ratio[~self.df_ratio.index.isin(self._test_df.index)]
        
        return self._test_df
        
    
    def subsample(self, subsample_factor, random_state=101, raise_exception=False):
        npr.seed(random_state)
        # subsample each label category inidividually to keep intrinsic ratios
        
        # use one hot encoding to calculate the distribution of the categorical features
        if len(self.categorical_columns):
            df_ratio_one_hot_encoded = self.df_ratio.drop(self.categorical_columns, axis=1).join(pd.get_dummies(self.df_ratio[self.categorical_columns], prefix_sep=self._column_separator))
        else:
            df_ratio_one_hot_encoded = self.df_ratio
        df_subsampled = df_ratio_one_hot_encoded.sample(frac=subsample_factor, replace=False, random_state=random_state)
        
        if self.allow_zero_occurrence and len(self.categorical_columns):
            # relevant only for categorical columns
            for col in df_subsampled:
                if col not in self.df_ratio:
                    # categorical columns have appendix based on value and are thus not in the columns_keep_ratio list
                    if not df_subsampled[col].any():
                        column, value = col.split(self._column_separator)
                        message = f'Could not find subsample with seed {random_state} due to zero occcurrences of {value} in column {column}.'
                        if raise_exception:
                            raise Exception(message)
                        else:
                            logging.warning(message)
                            return None

        # check for allowed divergence
        deviation_list = self.calculate_deviations(df_subsampled)
        for deviation_dict in deviation_list:
            col, deviation = deviation_dict['col'], deviation_dict['deviation']
            if deviation > self.allowed_deviation:
                message = f'Could not find subsample with seed {random_state} due to mean deviation {deviation} in column {col}. Allowed deviation is set to {self.allowed_deviation}.'
                if raise_exception:
                    raise Exception(message)
                else:
                    logging.warning(message)
                    return None
            
        indices = set(df_subsampled.index)
        df_subsampled = self.df_data[self.df_data.index.isin(indices)]
        return df_subsampled
    
    
    def calculate_deviations(self, df_subsampled):
        deviation_list = []
        for col, col_mean in df_subsampled.mean().items():
            deviation = np.abs(self.df_mean_orig[col] - col_mean)
            deviation_list.append({'col': col, 'deviation': deviation})
        self._deviation_list = deviation_list
        return deviation_list
    
    
    def get_deviation_list(self):
        return self._deviation_list
    
    
    @staticmethod
    def column_normalize(col):
        col_min = col.min()
        col_max = col.max()
        if col_min == col_max == 0:
            # all 0s
            return col
        return (col - col_min) / (col_max - col_min)
    




import unittest
import dazer
import pandas as pd
import seaborn as sns
import numpy as np
import tempfile
import shutil
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

class Testing(unittest.TestCase):


    def setUp(self) -> None:
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        return super().setUp()


    def tearDown(self) -> None:
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)
        return super().tearDown()


    def test_Classifier_random_forest(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test(
            'rf', param_grid={
                'bootstrap': [True],
                'max_depth': [2],
                'class_weight': ['balanced'],
                'min_samples_split': [2],
                'min_samples_leaf': [1],
                'n_estimators': [10],
                'random_state': [101]
            }, scoring='f1', cv=3)
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.7879)
        
        
    def test_Classifier_xgboost(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test(
            'xgb', scoring='f1')
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.8182)
        
        
    def test_Classifier_mlp(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test(
            'mlp', scoring='f1')
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.8485)
        
        
    def test_Classifier_mlp_parameters(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(
            X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test('mlp', scoring='f1', param_model={
                                                  'solver': 'adam', 'hidden_layer_sizes': (2, 1), 'this_parameter_does_not_exist': None})
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.4545)


    def test_Classifier_gnb(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(
            X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test('gnb', scoring='f1')
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.8788)
    
    
    def test_Classifier_svc(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(
            X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test('svc', scoring='f1')
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.4545)
        
        
    def test_Classifier_svc_parameters(self):
        X, y = make_classification(
            n_samples=100, n_features=10, random_state=444)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.33, random_state=42)
        classifier = dazer.Classifier(pd.DataFrame(
            X_train), y_train, pd.DataFrame(X_test), y_test)
        model, evaluation = classifier.train_test('svc', scoring='f1', param_model={
                                                  'kernel': 'linear', 'C':0.025, 'this_parameter_does_not_exist': None})
        self.assertTrue(round(evaluation['accuracy'], 4) == 0.8788)



import joblib
import numpy.random as npr
import numpy as np
import dazer


def subsample_iterative(df, columns_keep_ratio=[], allowed_deviation=.2, test_size=.2, random_states=[101, 102, 103, 104, 105], attempts=10000, ratios=np.arange(0, 1, .1)):
    """
     Iteratively subsamples a data frame to test data. The subsample is done by randomly choosing a subset of columns from the data frame and then re - sampling the data with a variety of random numbers.
     
     Args:
     	 df: Data frame to be subsampled.
     	 columns_keep_ratio: List of column names to keep ratio is ignored.
     	 allowed_deviation: Allowed deviation is used to determine the probability of a column being considered to be significant.
     	 test_size: Number of samples to be tested.
     	 random_states: List of random numbers to use for the subsampling.
     	 attempts: Number of attempts to subsample before giving up.
     	 ratios: Numpy array of ratios to use for each subsample.
     
     Returns: 
     	 Two dicts with DataFrame values, the first one containing test data and the second one training data. The keys containg information about ratios and random seeds. If there is an error in the data a ValueError is raised
    """
    df_test_dict = {}
    df_train_dict = {}
    # Generate a random state for each of the random states.
    for random_state in random_states:
        subsampler = dazer.Subsampler(
            df, columns_keep_ratio=columns_keep_ratio, allowed_deviation=allowed_deviation)
        npr.seed(random_state)
        i = 0
        # This function will attempt to extract test data from the dataset.
        while i < attempts:
            random_state_subsample = npr.randint(1, 999999999)
            try:
                df_test = subsampler.extract_test(
                    test_size=test_size, random_state=random_state_subsample)
                # Check if df_test is None.
                if df_test is None:
                    raise Exception()
                df_test_dict[f'{random_state}_{random_state_subsample}'] = df_test
                break
            except Exception as e:
                print(e)
                i += 1
        
        # Check if df_test is None.
        if df_test is None:
            raise Exception()
                    
        # This function generates a number of random samples for each ratio.
        for ratio in ratios:
            # If ratio 0 continue to do so.
            if ratio == 0:
                continue
            ratio = round(ratio, 4)
            
            npr.seed(random_state)
            i = 0
            # This function will generate a 'attempts' number of times a random number of attempts.
            while i < attempts:
                random_state_subsample = npr.randint(1, 999999999)
                try:
                    df_train_dict[f'{random_state}_{random_state_subsample}_{ratio}'] = subsampler.subsample(subsample_factor=ratio, random_state=random_state_subsample, raise_exception=True)
                    break
                except Exception as e:
                    print(e)
                    i += 1
    return df_test_dict, df_train_dict


def load_models(paths):
    """
     Load models from a list of paths. This is a convenience function for use with joblib. load ()
     
     Args:
     	 paths: A list of paths to load models from.
     
     Returns: 
     	 A list of : class : list objects corresponding to the loaded models.
    """
    models = []
    # Load jobs from paths.
    for path in paths:
        models.append(joblib.load(path))
    return models


def evaluate_prediction(y, y_pred):
    """
     Evaluate the prediction. This is a helper function for predict_proba. It takes the y values and the predicted values and returns a dictionary that maps the values to their confidence scores.
     
     Args:
     	 y: A list of values to be used for evaluation.
     	 y_pred: A list of predicted values. Each value should be 1 or 0.
     
     Returns: 
     	 A dictionary with the following keys : tp fp fn n_tp n_fp n_fn tnr fn
    """
    n_tp = 0
    n_tn = 0
    n_fp = 0
    n_fn = 0
    for iy, iy_pred in zip(y, y_pred):
        if iy == iy_pred:
            if iy == 1:
                n_tp += 1
            else:
                n_tn += 1
        else:
            if iy == 1:
                n_fn += 1
            else:
                n_fp += 1
    
    tpr = n_tp/(n_tp + n_fp) if (n_tp + n_fp) > 0 else 0
    fpr = n_fp/(n_fp + n_tn) if (n_fp + n_tn) > 0 else 0
    tnr = n_tn/(n_tn + n_fp) if (n_tn + n_fp) > 0 else 0
    fnr = n_fn/(n_fn + n_tn) if (n_fn + n_tn) > 0 else 0
    return {'tp': tpr, 'fp': fpr, 'tn': tnr, 'fn': fnr, 'n_tp': n_tp, 'n_tn': n_tn, 'n_fn': n_fn, 'n_fp': n_fp}
        
        
def random_classifier_prediction_evaluation(n_models, y_test, n_X, random_state=101):
    """
     Generate a list of predictions for each model. This is useful for testing the performance of different classifiers
     
     Args:
     	 n_models: Number of models to generate
     	 y_test: Test labels for each model
     	 n_X: Number of data points to use in evaluation
     	 random_state: Random state to use for npr. seed
     
     Returns: 
     	 List of evaluations for each model ( prediction_evaluation_list ). Each evaluation is a list
    """
    npr.seed(random_state)
    prediction_evaluation_list = []
    # This function evaluates the prediction of the model.
    for _ in range(n_models):
        y_random = [npr.random() > 0.5 for _ in range(n_X)]
        evaluation = evaluate_prediction(y_test, y_random)
        prediction_evaluation_list.append(evaluation)
    return prediction_evaluation_list



from dazer import utils


def random_forests_feature_importances(models):
    feature_importances = []
    
    for model in models:
        importances = model.best_estimator_.feature_importances_.tolist()
        feature_importances.append(importances)

    return feature_importances

def random_forests_feature_importances_from_files(paths):
    models = utils.load_models(paths)
    feature_importances = random_forests_feature_importances(models)
    return feature_importances



import unittest
from dazer import Subsampler 
import pandas as pd
import seaborn as sns
import numpy as np
import tempfile
import shutil

class Testing(unittest.TestCase):
    
    def setUp(self) -> None:
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        return super().setUp()
    
    
    def tearDown(self) -> None:
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)
        return super().tearDown()
    
    
    def test_column_normalize(self):
        col = pd.Series([-0.25, 3])
        col_norm = Subsampler.column_normalize(col)
        self.assertTrue(col_norm.min() == 0)
        self.assertTrue(col_norm.max() == 1)
    
    
    def test_Subsampler(self):
        
        df = sns.load_dataset('penguins', data_home=self.test_dir)
        subsampler = Subsampler(df, ['body_mass_g'], .07, True)
        
        df_test = subsampler.extract_test(.2, random_state=2)
        df_train = subsampler.subsample(.4, random_state=3)
        
        deviation_test = np.abs(subsampler.column_normalize(df['body_mass_g']).mean() - subsampler.column_normalize(df_test['body_mass_g']).mean())
        self.assertTrue(round(deviation_test, 5) == round(0.028691008567169607, 5))
            
        deviation_train = np.abs(subsampler.column_normalize(df['body_mass_g']).mean() - subsampler.column_normalize(df_train['body_mass_g']).mean())
        self.assertTrue(round(deviation_train, 5) == round(0.010262500500191507, 5))
        
        
    def test_subsampling_deviation(self):
        df = sns.load_dataset('penguins', data_home=self.test_dir)
        subsampler = Subsampler(df, ['body_mass_g'], .02, True)
        
        df_test = subsampler.extract_test(.2, random_state=2)
        self.assertTrue(df_test is None)
    
        
    def test_subsampling_get_deviation_list(self):
        df = sns.load_dataset('penguins', data_home=self.test_dir)
        subsampler = Subsampler(df, ['body_mass_g'], .06, True)
        
        _ = subsampler.extract_test(.4, random_state=4)
        _ = subsampler.subsample(.4, random_state=4)
        
        devialtion_list = subsampler.get_deviation_list()
        
        self.assertTrue(round(devialtion_list[0]['deviation'], 5) == round(0.01643383149231098, 5))
        
        
    def test_subsampling_allow_zero_occurrence(self):
        df = sns.load_dataset('penguins', data_home=self.test_dir)
        subsampler = Subsampler(df, ['body_mass_g', 'island', 'species'], 1, True)
        
        df_train = subsampler.subsample(.01, random_state=4)
        
        self.assertTrue(df_train is None)




"""
dazer, dataset size effect estimator
"""

__version__ = "0.1.21"
__author__ = 'Michael Hartung'
__credits__ = """Michael Hartung, Institute for Computational Systems Biology (https://www.cosy.bio/), University of Hamburg"""

from .Subsampler import Subsampler
from .Classifier import Classifier
from .utils import load_models
from . import defaults
from .random_forest_utils import random_forests_feature_importances_from_files, random_forests_feature_importances



class ParameterGrid:
    random_forest = {
                'bootstrap': [True],
                'max_depth': [1, 2, 5, 10, 50, None],
                'class_weight': ['balanced'],
                'min_samples_split': [2, 4, 8],
                'min_samples_leaf': [1, 2, 4, 8],
                'n_estimators': [10, 100, 250, 500, 750, 1000],
                'random_state': [101]
            }
    
    


from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import classification_report
from pathlib import Path
import joblib
import os
# from sklearn.inspection import permutation_importance
import pandas as pd
from dazer import utils


class Classifier:
    
    def __init__(self, X_train, y_train, X_test, y_test):
       
        self.label_encodings = {}
        
        X_test = X_test.copy()
        X_train = X_train.copy()
        
        # encode categorical columns
        categorical_columns = X_train.columns[~X_test.columns.isin(X_test._get_numeric_data().columns)]
        for col in categorical_columns:
            self.label_encodings[col] = {x: i for i, x in enumerate(X_test[col].unique())}
            X_test[col] = X_test[col].map(self.label_encodings[col])
            X_train[col] = X_train[col].map(self.label_encodings[col])
            
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test
        
        
    @staticmethod
    def get_model_from_string(model_string: str, random_state=101, hidden_layer_sizes=(10, 5), solver='lbfgs', alpha=1e-5, kernel='rbf', C=1, gamma=2, **_):
        if model_string == 'rf':
            return RandomForestClassifier()
        elif model_string == 'xgb':
            from xgboost import XGBClassifier
            return XGBClassifier(objective='binary:logistic')
        elif model_string == 'mlp':
            from sklearn.neural_network import MLPClassifier
            return MLPClassifier(solver=solver, alpha=alpha, hidden_layer_sizes=hidden_layer_sizes, random_state=random_state)
        elif model_string == 'gnb':
            from sklearn.naive_bayes import GaussianNB
            return GaussianNB()
        elif model_string == 'svc':
            from sklearn.svm import SVC
            return SVC(kernel=kernel, gamma=gamma, C=C, random_state=random_state)
            
    
    def train_test(self, model_string: str, param_model={},  param_grid={}, cv=10, n_jobs=-1, model_path='', verbose=1, scoring='f1'):
        classifier = self.get_model_from_string(model_string, **param_model)
        model = self.train(classifier, param_grid=param_grid, cv=cv,
                           n_jobs=n_jobs, verbose=verbose, scoring=scoring)
        
        y_pred = self.test(model)
        evals = self.eval_pred(y_pred)
        
        if model_path:
            Path(os.sep.join(model_path.split(os.sep)[:-1])).mkdir(parents=True, exist_ok=True)
            joblib.dump(model, model_path)

        return model, evals
    
    
    def train(self, model, param_grid={}, cv=10, n_jobs=-1, verbose = 1, scoring='f1'):
        model = GridSearchCV(estimator = model, param_grid = param_grid, cv = cv, n_jobs = n_jobs, verbose = verbose, scoring=scoring)
        model.fit(self.X_train, self.y_train)
        return model
    
    
    def test(self, model):
        return model.predict(self.X_test)
    
    
    def eval_pred(self, y_pred):
       clrep = classification_report(self.y_test, y_pred, target_names=None, output_dict=True)
       return {'n_samples_train': len(self.X_train), 
                'n_samples_test': len(self.X_test), 
                'accuracy': clrep['accuracy'],
                'f1': clrep['1']['f1-score'],
                'precision': clrep['1']['precision'],
                'recall': clrep['1']['recall'],
                'TNR': clrep['0']['recall'],
                }
        
        
    def get_feature_weights_random_forests(self, model_paths=[]):
        weights = []
        columns = self.X_train.columns
        for rf_path in  model_paths:
            clf = joblib.load(rf_path)
            data = list(clf.best_estimator_.feature_importances_)
            data.append(rf_path)
            weights.append(data)
            
        columns.append('model_path')
        df = pd.DataFrame(data=weights)

        df['model'] = df['model_path'].map(lambda x: x.split('/')[-1])
        return df
    

    def classifier_prediction_evaluation(self, models):
        prediction_evaluation_list = []
        for model in models:
            y_pred = model.predict(self.X_test)
            evaluation = utils.evaluate_prediction(self.y_test, y_pred)
            prediction_evaluation_list.append(evaluation)
        return prediction_evaluation_list
    
    
    def simulate_random_classifier_prediction_evaluation(self, n_models, random_state=101):
        
        random_evaluation_list = utils.random_classifier_prediction_evaluation(n_models, self.y_test, len(self.X_test), random_state)
        
        return random_evaluation_list
        
    
    
    # def permutation_test_random_forest(self, model_path, ratio, random_state=101):
    #     # random forest
    #     model = joblib.load(model_path)
        
    #     permutation_result = permutation_importance(
    #         model, self.X_train, self.y_train, n_repeats=10, random_state=random_state, n_jobs=5
    #     )
        
    #     return {
    #         'ratio': ratio,
    #         'dataset': self.dataset_name,
    #         'permutation_importances_mean': permutation_result.importances_mean,
    #         'permutation_importances_std': permutation_result.importances_std
    #     }


from setuptools import setup

setup(
    name='dazer',
    version='0.1.21',
    author='Maiykol',
    author_email='michael.hartung@uni-hamburg.de',
    packages=['dazer'],
    scripts=[],
    url='http://pypi.python.org/pypi/dazer/',
    license='LICENSE',
    description='DAtaset siZe Effect estimatoR',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "pandas==2.1.1",
        "scikit-learn==1.3.1",
        "xgboost==2.0.2"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        # 'License :: OSI Approved :: BSD License',  
        # 'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3.11',
    ],
)